#!/usr/bin/python

# Copyright 2009 Canonical Ltd.

"""Backport a source package from another release."""

import sys
import os
from optparse import OptionParser
import urllib2

sys.path.insert(0, os.path.join(
    os.path.dirname(__file__), 'lib'))

from debian_bundle import deb822
import apt_pkg

import lputils
import synclib

CONSUMER_KEY = "backport.py"


def find_dsc(options, pkg):
    fromname = '%s-%s' % (options.fromseries.name, options.frompocket.lower())

    spp = []
    spp.extend([('from', sp)
                for sp in options.fromdistro.main_archive.getPublishedSources(
                    distro_series=options.fromseries,
                    pocket=options.frompocket, source_name=pkg,
                    exact_match=True, status='Published')])
    spp.extend([('to', sp)
                for sp in options.todistro.main_archive.getPublishedSources(
                    distro_series=options.toseries,
                    source_name=pkg,
                    exact_match=True, status='Published')])

    latest = {'from': None, 'to': None}
    pkgs = {'from': {}, 'to': {}}
    for suite, sp in spp:
        name = sp.source_package_name
        version = sp.source_package_version
        if (latest[suite] is not None and
            apt_pkg.VersionCompare(latest[suite], version) >= 0):
            continue
        latest[suite] = version
        # sp.files not exposed via the API yet; see
        # https://bugs.launchpad.net/bugs/394827
        changes_handle = urllib2.urlopen(sp.changesFileUrl())
        try:
            changes = deb822.Changes(changes_handle.read())
        finally:
            changes_handle.close()
        for sp_file in changes['files']:
            if sp_file['name'].endswith('.dsc'):
                if name.startswith('lib') and len(name) > 3:
                    name_hash = name[:4]
                else:
                    name_hash = name[0]
                dirname = 'pool/%s/%s/%s' % (sp.component_name, name_hash, name)
                pkgs[suite] = (sp, '%s/%s' % (dirname, sp_file['name']))
                break

    if latest['to'] is not None:
        if apt_pkg.VersionCompare(latest['from'], latest['to']) == 0:
            print >>sys.stderr, ("%s is in both %s and %s with the same version (%s) - skipping." %
                                 pkg, fromname, options.toseries.name,
                                 latest['from'])
            return None, None

        if apt_pkg.VersionCompare(latest['from'], latest['to']) < 0:
            print >>sys.stderr, ("%s is older in %s (%s) than %s (%s) ... say what now? - skipping." %
                                 pkg, fromname, latest['from'],
                                 options.toseries.name, latest['to'])
            return None, None

    return pkgs['from'][1], latest['to']


def backport(options, pkg, failures):
    print " * Trying to backport %s..." % pkg

    dsc_filename, previous_version = find_dsc(options, pkg)

    extra_opts = [options.requestor, options.toseries.name,
                  dsc_filename, previous_version]
    retval, output = synclib.do_sync('backport', pkg, extra_opts)
    if not retval or not output:
        failures.append('backport-source failed')


def parse_options(args=None):
    parser = OptionParser(usage="Usage: %prog [options] package ...")

    parser.add_option("-l", "--launchpad", dest="launchpad_instance",
                      default="edge")

    parser.add_option("-b", "--requested-by", dest="requestor",
                      help="who the backport was requested by")

    parser.add_option("-d", "--to-distro", dest="todistro",
                      help="backport to DISTRO",
                      default="ubuntu")
    parser.add_option("-s", "--to-suite", dest="tosuite",
                      help="backport to SUITE")

    parser.add_option("-D", "--from-distro", dest="fromdistro",
                      help="backport from DISTRO",
                      default="ubuntu")
    parser.add_option("-S", "--from-suite", dest="fromsuite",
                      help="backport from SUITE")

    options, args = parser.parse_args(args=args)
    if len(args) < 1:
        parser.error("need at least one source package name")

    return options, args


def default_options(launchpad, options):
    if not options.fromdistro:
        options.fromdistro = 'ubuntu'
    options.fromdistro = launchpad.distributions[options.fromdistro]
    if not options.fromsuite:
        options.fromseries = options.fromdistro.current_series
        options.frompocket = 'Release'
    elif '-' in options.fromsuite:
        options.fromseries, options.frompocket = options.fromsuite.split('-', 1)
        options.fromseries = options.fromdistro.getSeries(
            name_or_version=options.fromseries)
        options.frompocket = options.frompocket.title()
    else:
        options.fromseries = options.fromdistro.getSeries(
            name_or_version=options.fromsuite)
        options.frompocket = 'Release'

    options.todistro = launchpad.distributions[options.todistro]
    options.toseries = options.todistro.getSeries(
        name_or_version=options.tosuite)

    if options.requestor:
        person = launchpad.people[options.requestor]
        try:
            options.requestor = '%s <%s>' % (person.display_name,
                                             person.preferred_email_address.email)
        except:
            # https://launchpad.net/bugs/319432
            # fails if email address is hidden
            print >>sys.stderr, ("Couldn't set requestor for '%s'. Using 'Ubuntu Archive Backport'" % (options.requestor))
            options.requestor = 'Ubuntu Archive Backport <archive@ubuntu.com>'
    else:
        options.requestor = 'Ubuntu Archive Auto-Backport <archive@ubuntu.com>'


def main():
    apt_pkg.init()

    options, args = parse_options()

    launchpad = lputils.lpfactory(options.launchpad_instance, CONSUMER_KEY)

    default_options(launchpad, options)

    failures = []
    for pkg in args:
        backport(options, pkg, failures)

    if failures:
        print >>sys.stderr, '********** FAILED BACKPORTS **********'
        print >>sys.stderr, '\n'.join(failures)
        sys.exit(1)

    print >>sys.stderr, "Remember to run mass-sync.py --flush-backports."

if __name__ == '__main__':
    main()
