#!/usr/bin/python

# Copyright 2008, 2009 Canonical Ltd.  All rights reserved.

"""Edit uploader permissions for the Ubuntu distro in Launchpad."""

import sys
import os
from optparse import OptionParser

sys.path.insert(0, os.path.join(
    os.path.dirname(__file__), 'lib'))

import launchpadlib.errors

import lputils


CONSUMER_KEY = "edit_acl.py"


def print_perms(perms, series=None):
    for perm in perms:
        if (series is not None and perm.distro_series_name is not None and
            series != perm.distro_series_name):
            continue
        desc = []
        desc.append("archive '%s'" % perm.archive.name)
        if perm.component_name:
            desc.append("component '%s'" % perm.component_name)
            if series:
                desc[-1] += ' in %s' % series
        if perm.package_set_name:
            desc.append("package set '%s' in %s" % (perm.package_set_name,
                                                    perm.distro_series_name))
        if perm.source_package_name:
            desc.append("source package '%s'" % perm.source_package_name)
        print "%s for %s: %s" % (perm.permission, perm.person.name,
                                 ', '.join(desc))


def get_archive(options, launchpad):
    distros = launchpad.distributions
    ubuntu = distros['ubuntu']
    if options.archive is None:
        return ubuntu.main_archive
    else:
        for archive in ubuntu.archives:
            if archive.name == options.archive:
                return archive
    raise AssertionError("No such archive in Ubuntu: %s" % options.archive)


def get_source_components(options, launchpad, archive, source):
    from debian_bundle import debian_support

    args = {}
    if options.series:
        args['distro_series'] = launchpad.distributions['ubuntu'].getSeries(
            name_or_version=options.series)

    newest = {}
    for spph in archive.getPublishedSources(
            source_name=source, exact_match=True, status='Published', **args):
        if not spph.distro_series.active:
            continue
        new_version = debian_support.Version(spph.source_package_version)
        if (spph.distro_series.name not in newest or
            new_version > newest[spph.distro_series.name][0]):
            newest[spph.distro_series.name] = (new_version,
                                               spph.component_name)

    for series in sorted(newest.keys(), key=lambda s: newest[s][0]):
        yield series, newest[series][1]


permission_names = dict(upload='Archive Upload Rights',
                        admin='Queue Administration Rights')


def do_query(options):
    """Query existing permissions and show on stdout."""
    launchpad = lputils.lpfactory(options.launchpad_instance, CONSUMER_KEY)
    if options.archive is None:
        archives = launchpad.distributions['ubuntu'].archives
    else:
        archives = [get_archive(options, launchpad)]

    if options.person:
        if '@' in options.person:
            person = launchpad.people.getByEmail(email=options.person)
        else:
            try:
                person = launchpad.people[options.person]
            except KeyError:
                print "Person '%s' doesn't exist." % options.person
                sys.exit(1)
        perms = []
        for archive in archives:
            perms.extend(archive.getPermissionsForPerson(person=person))
        if options.acl_type:
            perm_name = permission_names[options.acl_type]
            perms = [p for p in perms if p.permission == perm_name]
        print "== All rights for %s ==" % person.name
        print_perms(perms, options.series)

    if options.component:
        perms = []
        if not options.acl_type or options.acl_type == 'upload':
            for archive in archives:
                perms.extend(archive.getUploadersForComponent(
                    component_name=options.component))
        if not options.acl_type or options.acl_type == 'admin':
            for archive in archives:
                perms.extend(archive.getQueueAdminsForComponent(
                    component_name=options.component))
        print "== All rights for component '%s' ==" % options.component
        print_perms(perms, options.series)

    if options.packageset:
        lp_series = launchpad.distributions['ubuntu'].getSeries(
            name_or_version=options.series)
        for packageset in options.packageset:
            lp_set = launchpad.packagesets.getByName(name=packageset,
                                                     distroseries=lp_series)

            perms = []
            for archive in archives:
                perms.extend(archive.getUploadersForPackageset(
                    packageset=lp_set))
            print ("== All uploaders for package set '%s' in '%s' ==" %
                   (packageset, options.series))
            print_perms(perms, options.series)

            sources = sorted(lp_set.getSourcesIncluded(direct_inclusion=True))
            if sources:
                print
                print ("== All source packages in package set '%s' "
                       "in '%s' ==" % (packageset, options.series))
                for source in sources:
                    print source
            child_sets = list(lp_set.setsIncluded(direct_inclusion=True))
            if child_sets:
                print
                print ("== All package sets in package set '%s' in '%s' ==" %
                       (packageset, options.series))
                for child_set in child_sets:
                    print child_set.name

    if options.source:
        for source in options.source:
            perms = []
            perms_set = []
            for archive in archives:
                perms.extend(archive.getUploadersForPackage(
                    source_package_name=source))
                perms_set.extend(archive.getPackagesetsForSource(
                    sourcepackagename=source))
            print "== All uploaders for package '%s' ==" % source
            print_perms(perms, options.series)
            print_perms(perms_set, options.series)
            for archive in archives:
                for series, component in get_source_components(
                        options, launchpad, archive, source):
                    perms_component = archive.getUploadersForComponent(
                        component_name=component)
                    print_perms(perms_component, series=series)


def validate_add_delete_options(options):
    if options.packageset and options.source:
        # Special options to manage package sets, bodged into this tool
        # since they aren't entirely inconvenient here.
        if options.component or options.person:
            print ("-P <packageset> -s <source> cannot be used with a "
                   "component or person as well")
            return False
        return True

    if not options.person:
        print "You must specify a person to authorise."
        return False

    count = 0
    if options.component:
        count += 1
    if options.packageset:
        count += 1
    if options.source:
        count += 1
    if count > 1:
        print "You can only specify one of package set, source, or component"
        return False

    if count == 0:
        print "You must specify one of package set, source, or component"
        return False

    return True


def do_add(options):
    """Add a new permission."""
    if not validate_add_delete_options(options):
        return False

    launchpad = lputils.lpfactory(options.launchpad_instance, CONSUMER_KEY)
    archive = get_archive(options, launchpad)

    if options.packageset and options.source:
        lp_series = launchpad.distributions['ubuntu'].getSeries(
            name_or_version=options.series)
        for packageset in options.packageset:
            lp_set = launchpad.packagesets.getByName(name=packageset,
                                                     distroseries=lp_series)
            lp_set.addSources(names=options.source)
            print "Added:"
            for source in options.source:
                print source
        return

    person = launchpad.people[options.person]

    if options.source:
        for source in options.source:
            perm = archive.newPackageUploader(
                person=person, source_package_name=source)
            print "Added:"
            print_perms([perm,])
        return

    if options.packageset:
        lp_series = launchpad.distributions['ubuntu'].getSeries(
            name_or_version=options.series)
        for packageset in options.packageset:
            lp_set = launchpad.packagesets.getByName(name=packageset,
                                                     distroseries=lp_series)
            perm = archive.newPackagesetUploader(
                person=person, packageset=lp_set)
            print "Added:"
            print_perms([perm,])
        return

    if options.component:
        if not options.acl_type or options.acl_type == 'upload':
            perm = archive.newComponentUploader(
                person=person, component_name=options.component)
        else:
            perm = archive.newQueueAdmin(
                person=person, component_name=options.component)
        print "Added:"
        print_perms([perm,])
        return


def do_delete(options):
    """Delete a permission."""
    if not validate_add_delete_options(options):
        return False

    launchpad = lputils.lpfactory(options.launchpad_instance, CONSUMER_KEY)
    archive = get_archive(options, launchpad)

    if options.packageset and options.source:
        lp_series = launchpad.distributions['ubuntu'].getSeries(
            name_or_version=options.series)
        for packageset in options.packageset:
            lp_set = launchpad.packagesets.getByName(name=packageset,
                                                     distroseries=lp_series)
            lp_set.removeSources(names=options.source)
            print "Deleted:"
            for source in options.source:
                print source
        return

    person = launchpad.people[options.person]

    if options.source:
        for source in options.source:
            archive.deletePackageUploader(
                person=person, source_package_name=source)
            print "Deleted %s/%s" % (options.person, source)
        return

    if options.packageset:
        lp_series = launchpad.distributions['ubuntu'].getSeries(
            name_or_version=options.series)
        for packageset in options.packageset:
            lp_set = launchpad.packagesets.getByName(name=packageset,
                                                     distroseries=lp_series)
            archive.deletePackagesetUploader(person=person, packageset=lp_set)
            print "Deleted %s/%s/%s" % (options.person, packageset,
                                        options.series)
        return

    if options.component:
        if not options.acl_type or options.acl_type == 'upload':
            archive.deleteComponentUploader(
                person=person, component_name=options.component)
            print "Deleted %s/%s" % (options.person, options.component)
        else:
            archive.deleteQueueAdmin(
                person=person, component_name=options.component)
            print "Deleted %s/%s (admin)" % (options.person, options.component)
        return


def do_create(options):
    if not options.packageset:
        print "You can only create a package set, not something else."
        return False

    launchpad = lputils.lpfactory(options.launchpad_instance, CONSUMER_KEY)
    techboard = launchpad.people['techboard']

    for packageset in options.packageset:
        try:
            if launchpad.packagesets.getByName(name=packageset):
                print "Package set %s already exists" % packageset
                continue
        except (TypeError, launchpadlib.errors.HTTPError):
            pass
        desc = raw_input("Description for new package set %s: " % packageset)
        ps = launchpad.packagesets.new(name=packageset, description=desc,
                                       owner=techboard)
        print ps


def main(options, action):

    if action == "query":
        do_query(options)
    elif action == "add":
        do_add(options)
    elif action in ("delete", "remove"):
        do_delete(options)
    elif action == "create":
        do_create(options)
    else:
        raise AssertionError("Invalid action %s" % action)

    return


if __name__ == '__main__':
    parser = OptionParser(usage="usage: %prog [options] query|add|delete|create")

    parser.add_option(
        "-l", "--launchpad", dest="launchpad_instance", default="edge")
    parser.add_option("-A", "--archive", dest="archive")
    parser.add_option("-p", "--person", dest="person")
    parser.add_option("-c", "--component", dest="component")
    parser.add_option("-P", "--packageset", dest="packageset", action="append")
    parser.add_option("-S", "--series", dest="series")
    parser.add_option("-s", "--source", dest="source", action="append")
    parser.add_option("-t", "--acl-type", dest="acl_type",
                      help="ACL type: upload or admin")

    options, args = parser.parse_args()

    possible_actions = ('query', 'add', 'delete')

    if len(args) != 1:
        parser.error(
            "You must specify an action, one of:\n%s"
                % ", ".join(possible_actions))

    if (not options.person and not options.component and
        not options.packageset and not options.source):
        parser.error("Provide at least one of "
                     "person/component/packageset/source")
    if options.packageset and not options.series:
        parser.error("Package set requires an associated series")
    if options.acl_type and options.acl_type not in ('upload', 'admin'):
        parser.error("Invalid ACL type '%s' (valid: 'upload', 'admin')" %
                     options.acl_type)
    if options.acl_type == 'admin' and options.packageset:
        parser.error("ACL type admin not allowed for package sets")
    if options.acl_type == 'admin' and options.source:
        parser.error("ACL type admin not allowed for source packages")

    try:
        main(options, args[0])
    except launchpadlib.errors.HTTPError, err:
        print "There was an error:"
        print err.content
