#!/usr/bin/python

# Copyright 2009 Canonical Ltd.

"""Process a list of bugs and optional extra sync/backport options from
stdin (one per line). The first word on each line needs to say whether to
perform a sync or a backport.

For each sync bug, do sync-source.py on all source package tasks with the
requestor being set to the bug reporter, and close the bug with a comment if
sync-source.py succeeded. Print out a list of failed syncs in the end.

For backports, the target suite will automatically be detected, unless you
use the -s option. Backports are a bit odd because the backport requests are
filed on SUITE-backports products rather than on source packages in a
distribution, which means we can't automatically detect which source package
to sync. Instead, you must specify this after the bug number.

You may similarly supply a source package in the case of syncs if necessary;
this should normally only be in cases of new packages, where the bug can't
yet be filed against the intended source package name.

Example:

  (echo sync 179263; echo sync 183336 -f; echo sync 183483 -S experimental;
   echo backport 224942 lintian) | ./mass-sync.py"""

import sys
import os
import optparse
from optparse import OptionParser
import urlparse
import shlex

sys.path.insert(0, os.path.join(
    os.path.dirname(__file__), 'lib'))

import apt_pkg

import lputils
import synclib
import backport

CONSUMER_KEY = "mass-sync.py"


def copy_options(options):
    return optparse.Values(options.__dict__)


def prepare_backport(launchpad, pkg, requestor, opts):
    options = copy_options(opts)
    options.requestor = requestor
    backport.default_options(launchpad, options)
    dsc_filename, previous_version = backport.find_dsc(options, pkg)
    return (options.requestor, options.toseries.name, dsc_filename,
            previous_version)


# Marshal parsed options back into a list of strings for the benefit of
# sync-source.py.
def prepare_sync(requestor, options):
    ret = ['-b', requestor]
    if options.force:
        ret.append('-f')
    if options.forcemore:
        ret.append('-F')
    if options.todistro != 'ubuntu':
        ret.append('-d')
        ret.append(options.todistro)
    if options.tosuite is not None:
        ret.append('-s')
        ret.append(options.tosuite)
    if options.fromcomponent is not None:
        ret.append('-C')
        ret.append(options.fromcomponent)
    if options.fromdistro is not None:
        ret.append('-D')
        ret.append(options.fromdistro)
    if options.fromsuite is not None:
        ret.append('-S')
        ret.append(options.fromsuite)
    if options.verbose is not None:
        ret.append('-V')
    return ret


def process_bug(launchpad, op_type, bugnum, source, options, failures):
    # Backporting may fiddle with options.tosuite, so save it.
    old_tosuite = options.tosuite
    # Similarly, for multi-task syncs, we need to clear the source every time
    old_source = source

    if op_type == 'backport':
        desc = 'backport-source'
    else:
        desc = 'sync-source'

    related_tasks = {}

    bug = launchpad.bugs[bugnum]
    synced = False
    for task in bug.bug_tasks:
        target = task.target
        target = launchpad.load(target.self_link)
        rtype = urlparse.urlsplit(target.resource_type_link)[4]
        if rtype == 'source_package':
            if (target.distroseries.name
                    == launchpad.distributions['ubuntu'].current_series.name):
                related_tasks[target.name] = 1

    for task in bug.bug_tasks:
        options.tosuite = old_tosuite
        source = old_source

        target = task.target
        # Horrible cheesy hack. This is basically a cast to the derived type
        # of target.
        # https://bugs.launchpad.net/bugs/340935
        target = launchpad.load(target.self_link)
        rtype = urlparse.urlsplit(target.resource_type_link)[4]
        if op_type == 'backport':
            assert source is not None
            if rtype != 'project':
                continue
            if options.tosuite:
                if target.name != ('%s-backports' % options.tosuite):
                    continue
            else:
                if not target.name.endswith('-backports'):
                    continue
                options.tosuite = target.name[:-10]
        elif source is None:
            if rtype == 'distribution':
                failures.append('%s - no source package given' % bugnum)
                continue
            if rtype != 'distribution_source_package':
                if (rtype != 'source_package'
                    or target.distroseries.name
                    != launchpad.distributions['ubuntu'].current_series.name):
                    print >>sys.stderr, "not a source package: %s" % rtype
                    continue
            if rtype == 'distribution_source_package':
                if target.name in related_tasks:
                    # It's got a series targetted task, so let that one be
                    # changed
                    continue
            if target.distribution.name != 'ubuntu':
                print >>sys.stderr, "not ubuntu: %s" % target.distribution.name
                continue
            source = target.name
        else:
            if rtype == 'distribution_source_package':
                if target.name != source:
                    continue
                if target.name in related_tasks:
                    # It's got a series targeted task
                    continue
            elif rtype == 'distribution':
                if target.name != 'ubuntu':
                    continue
            elif rtype == 'source_package':
                if (target.distroseries.name
                    != launchpad.distributions['ubuntu'].current_series.name):
                    continue
            else:
                continue

        requestor = bug.owner.name
        if options.requestor:
            requestor = options.requestor

        print ('processing bug %s, package %s, reporter %s' %
               (bugnum, source, requestor))
        if op_type == 'backport':
            extra_opts = prepare_backport(launchpad, source, requestor,
                                          options)
        else:
            extra_opts = prepare_sync(requestor, options)
        retval, output = synclib.do_sync(op_type, source, extra_opts)
        synced = True
        if retval and output:
            # per Jordan Mantha, use the bug title when adding new comments
            # to avoid bug mails with useless subjects
            bug.newMessage(content=''.join(output), subject=bug.title)
            task.status = "Fix Released"
            task.lp_save()
        else:
            if not output:
                failures.append('%s - %s failed' % (bugnum, desc))
            else:
                failures.append('%s - %s failed:\n%s' % (bugnum, desc, output))
    if not synced:
        failures.append("%s - %s: mass-sync.py couldn't work out what to do "
                "from the bug" % (bugnum, desc))


def main():
    apt_pkg.init()

    parser = OptionParser(usage="Usage: %prog [options]")
    parser.add_option("-l", "--launchpad", dest="launchpad_instance",
                      default="edge")
    parser.add_option("--flush-backports", dest="flush_backports",
                      default=False, action='store_true')
    parser.add_option("--flush-syncs", dest="flush_syncs",
                      default=False, action='store_true')
    parser.add_option("-v", "--verbose", dest="verbose",
                      default=False, action='store_true')
    options, args = parser.parse_args()

    if options.flush_backports or options.flush_syncs:
        if options.flush_backports:
            subp = synclib.request('flush-backports')
            for line in subp.stdout:
                print line,
            subp.wait()
        if options.flush_syncs:
            subp = synclib.request('flush-syncs')
            for line in subp.stdout:
                print line,
            subp.wait()
        sys.exit(0)

    launchpad = lputils.lpfactory(options.launchpad_instance, CONSUMER_KEY)

    failures = []

    for line in sys.stdin:
        if not line.strip():
            continue

        lineparser = OptionParser()
        # This OptionParser needs to be told about all the backport/sync
        # options that one might want to use in this context. It's OK to
        # leave out options that are irrelevant for bug-based syncing (sync
        # -a) or going to be filled out in other ways.

        lineparser.add_option("-f", "--force", dest="force",
                              default=False, action="store_true",
                              help="force sync over the top of Ubuntu changes")
        lineparser.add_option("-F", "--force-more", dest="forcemore",
                              default=False, action="store_true",
                              help="force sync even when components don't match")

        lineparser.add_option("-d", "--to-distro", dest="todistro",
                              help="sync/backport to DISTRO",
                              default="ubuntu")
        lineparser.add_option("-s", "--to-suite", dest="tosuite",
                              help="sync/backport to SUITE")

        lineparser.add_option("-C", "--from-component", dest="fromcomponent",
                              help="sync from COMPONENT")
        lineparser.add_option("-D", "--from-distro", dest="fromdistro",
                              help="sync/backport from DISTRO")
        lineparser.add_option("-S", "--from-suite", dest="fromsuite",
                              help="sync/backport from SUITE")
        lineparser.add_option("-b", "--requested-by", dest="requestor",
                              help="who the sync/backport was requested by")

        lineopts, lineargs = lineparser.parse_args(args=shlex.split(line))

        if len(lineargs) < 2 or lineargs[0] not in ('sync', 'backport'):
            lineparser.error("need sync/backport and a bug number")
        try:
            int(lineargs[1])
        except ValueError:
            lineparser.error("'%s' not a bug number" % lineargs[1])
        if lineargs[0] == 'backport' and len(lineargs) < 3:
            lineparser.error("backport needs a source package name")

        if len(lineargs) < 3:
            source = None
        else:
            source = lineargs[2]

        lineopts.verbose = options.verbose

        process_bug(launchpad, lineargs[0], lineargs[1], source, lineopts,
                    failures)

    if failures:
        print >>sys.stderr, '********** FAILED SYNCS **********'
        print >>sys.stderr, '\n'.join(failures)
        sys.exit(1)

    print >>sys.stderr, \
        "Remember to run mass-sync.py --flush-syncs and/or --flush-backports."

if __name__ == '__main__':
    main()
